from django.apps import apps
from django.conf import settings
import os, inspect
import logging
from ..util import log


##############################################################
###   Abstract Provider Base

class BaseProvider(object):
    '''
    Abstract base provider class.  Instances of this class are created by ProviderRun's constructor.

    Note that the app can only be inferred for templates in project apps (below settings.BASE_DIR).
    The app has to be inferred because mako creates templates internally during the render runtime,
    and I don't want to hack into Mako internals.
    '''
    DEFAULT_OPTIONS = {
        # the group this provider is part of.  this only matters when
        # the html page limits the providers that will be called with
        # ${ django_mako_plus.links(group="...") }
        'group': 'styles',
        # whether enabled (see "Dev vs. Prod" in the DMP docs)
        'enabled': True,
    }


    def __init__(self, provider_run, template, index, options):
        # the following are always set
        self.provider_run = provider_run    # the object in charge of this run of providers for the given template
        self.template = template            # Mako template object
        self.index = index                  # position of this provider in the list for this run
        self.options = options              # the combined options from the provider, its supers, and settings.py

        # the only time these remain None is if we can't infer the app. that happens when:
        #   1. the template was created from a string (and has no filename), or
        #   2. the template file is located outside of an app directory
        self.app_config = None              # AppConfig the template resides in, if possible to infer.
        self.template_ext = None            # Template filename extension
        self.template_relpath = None        # Template path, relative to app/templates/ and without extension, if possible to infer
        self.template_name = None           # Template filename without extension, if possible to infer
        if self.template.filename is not None:
            # try to infer the app
            fn_no_ext, self.template_ext = os.path.splitext(template.filename)
            relpath = os.path.relpath(fn_no_ext, settings.BASE_DIR)
            if not relpath.startswith('..'):  # can't infer reliably outside of project dir
                try:
                    path_parts = os.path.normpath(relpath).split(os.path.sep)
                    self.app_config = apps.get_app_config(path_parts[0])
                    self.template_relpath = '/'.join(path_parts[2:])
                    self.template_name = path_parts[-1]
                except LookupError: # template isn't under an app
                    pass

    def __repr__(self):
        return '<{}/{}:{}{}>'.format(
            self.app_config.name if self.app_config is not None else 'unknown',
            self.template_relpath if self.template_relpath is not None else self.template,
            self.__class__.__qualname__,
            '' if self.options['enabled'] else ' (disabled)',
        )

    @property
    def group(self):
        return self.options['group']

    def provide(self):
        '''
        Generate the content and do the work of this provider.
        Use self.write() to output content.
        '''
        pass

    def write(self, content):
        '''Writes content to the response'''
        # really just a redirect to the provider run
        self.provider_run.write(content)


    # in these next methods, the concept of "related providers" means the providers
    # of the same class type in the same position.
    # Suppose we have index.html inheriting from base.htm and we have three providers
    # listed in settings. We get six total providers:
    #
    #    base.htm   ->  JsContextProvider, CssLinkProvider, JsLinkProvider
    #       |
    #   index.html  ->  JsContextProvider, CssLinkProvider, JsLinkProvider
    #
    # In this example, the two JsContextProviders are "related providers",
    # the two CssLinkProviders are "related providers", and the two
    # JsLinkProviders are "related providers".

    def iter_related(self):
        '''
        Generator function that iterates this object's related providers,
        which includes this provider.
        '''
        for tpl in self.provider_run.templates:
            yield tpl.providers[self.index]

    def get_first(self):
        '''
        Returns the first provider in the related providers to this one.
        This is the provider instance for the base template (e.g. base.htm).
        This is useful when a provider class needs to do things at the start of a run.
        '''
        return self.provider_run.templates[0].providers[self.index]

    def is_first(self):
        '''
        Returns true if this provider is first to run among its related providers.
        This is the provider associated with the base template (e.g. base.htm).
        '''
        return self.provider_run.templates[0].template is self.template

    def get_last(self):
        '''
        Returns the last provider in the related providers to this one.
        This is the provider instance for the main template (e.g. index.html).
        This is useful when a provider class needs to do things at the end of a run.
        '''
        return self.provider_run.templates[-1].providers[self.index]

    def is_last(self):
        '''
        Returns true if this provider is last to run among its related providers.
        This is the provider associated with the main template (e.g. index.htm).
        '''
        return self.provider_run.templates[-1].template is self.template



from django.apps import apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.management.base import BaseCommand, CommandError
from django_mako_plus import LinkProvider
from django_mako_plus.template import create_mako_context
from django_mako_plus.provider.runner import ProviderRun
from django_mako_plus.management.mixins import DMPCommandMixIn
from mako.template import Template as MakoTemplate
import os
import os.path
from collections import OrderedDict



class Command(DMPCommandMixIn, BaseCommand):
    help = 'Removes compiled template cache folders in your DMP-enabled app directories.'
    can_import_settings = True

    def __init__(self, *args, running_inline=False, **kwargs):
        # special option when called from providers/webpack.py
        self.running_inline = running_inline
        super().__init__(*args, **kwargs)


    def add_arguments(self, parser):
        super().add_arguments(parser)

        parser.add_argument(
            '--overwrite',
            action='store_true',
            dest='overwrite',
            default=False,
            help='Overwrite existing __entry__.js if it exists'
        )
        parser.add_argument(
            '--single',
            type=str,
            metavar='FILENAME',
            help='Instead of per-app entry files, create a single file that includes the JS for all listed apps'
        )
        parser.add_argument(
            'appname',
            type=str,
            nargs='*',
            help='The name of one or more DMP apps. If omitted, all DMP apps are processed'
        )


    def handle(self, *args, **options):
        dmp = apps.get_app_config('django_mako_plus')
        self.options = options
        WebpackProviderRun.initialize_providers()
        for pci in WebpackProviderRun.CONTENT_PROVIDERS:
            if not issubclass(pci.cls, LinkProvider):
                raise ImproperlyConfigured('Invalid provider {} listed in {}: must extend django_mako_plus.LinkProvider'.format(pci.cls.__qualname__, WebpackProviderRun.SETTINGS_KEY))

        # ensure we have a base directory
        try:
            if not os.path.isdir(os.path.abspath(settings.BASE_DIR)):
                raise CommandError('Your settings.py BASE_DIR setting is not a valid directory.  Please check your settings.py file for the BASE_DIR variable.')
        except AttributeError:
            raise CommandError('Your settings.py file is missing the BASE_DIR setting.')

        # the apps to process
        enapps = []
        for appname in options.get('appname'):
            enapps.append(apps.get_app_config(appname))
        if len(enapps) == 0:
            enapps = dmp.get_registered_apps()

        # main runner for per-app files
        created = False
        if options.get('single') is None:
            for app in enapps:
                self.message('Searching `{}` app...'.format(app.name))
                filename = os.path.join(app.path, 'scripts', '__entry__.js')
                created = created or self.create_entry_file(filename, self.generate_script_map(app), [ app ])

        # main runner for one sitewide file
        else:
            script_map = {}
            for app in enapps:
                self.message('Searching `{}` app...'.format(app.name))
                script_map.update(self.generate_script_map(app))
            created = created or self.create_entry_file(options.get('single'), script_map, enapps)


    def create_entry_file(self, filename, script_map, enapps):
        '''Creates an entry file for the given script map'''
        if len(script_map) == 0:
            return

        # create the entry file
        template = MakoTemplate('''
<%! import os %>

// Contains links for ${ 'app' if len(enapps) == 1 else 'apps' }: ${ ', '.join(sorted([ a.name for a in enapps ])) }
(context => {
    DMP_CONTEXT.loadBundle({
      %for (app, template), script_paths in script_map.items():
        "${ app }/${ template }": function() {
          %for path in script_paths:
            require("./${ os.path.relpath(path, os.path.dirname(filename)) }");
          %endfor
        },
      %endfor
    });
})(DMP_CONTEXT.get());
''')
        content = template.render(
            enapps=enapps,
            script_map=script_map,
            filename=filename,
        ).strip()

        # ensure the parent directories exist
        if not os.path.exists(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename))

        # if the file exists, then consider the options
        file_exists = os.path.exists(filename)
        if file_exists and self.running_inline:
            # running inline means that we're in debug mode and webpack is likely watching, so
            # we don't want to recreate the entry file (and cause webpack to constantly reload)
            # unless we have changes
            with open(filename, 'r') as fin:
                if content == fin.read():
                    return False
        if file_exists and not self.options.get('overwrite'):
            raise CommandError('Refusing to destroy existing file: {} (use --overwrite option or remove the file)'.format(filename))

        # if we get here, write the file
        self.message('Creating {}'.format(os.path.relpath(filename, settings.BASE_DIR)))
        with open(filename, 'w') as fout:
            fout.write(content)
        return True

    def generate_script_map(self, config):
        '''
        Maps templates in this app to their scripts.  This function deep searches
        app/templates/* for the templates of this app.  Returns the following
        dictionary with absolute paths:

        {
            ( 'appname', 'template1' ): [ '/abs/path/to/scripts/template1.js', '/abs/path/to/scripts/supertemplate1.js' ],
            ( 'appname', 'template2' ): [ '/abs/path/to/scripts/template2.js', '/abs/path/to/scripts/supertemplate2.js', '/abs/path/to/scripts/supersuper2.js' ],
            ...
        }

        Any files or subdirectories starting with double-underscores (e.g. __dmpcache__) are skipped.
        '''
        script_map = OrderedDict()
        template_root = os.path.join(os.path.relpath(config.path, settings.BASE_DIR), 'templates')
        def recurse(folder):
            subdirs = []
            if os.path.exists(folder):
                for filename in os.listdir(folder):
                    if filename.startswith('__'):
                        continue
                    filerel = os.path.join(folder, filename)
                    if os.path.isdir(filerel):
                        subdirs.append(filerel)

                    elif os.path.isfile(filerel):
                        template_name = os.path.relpath(filerel, template_root)
                        scripts = self.template_scripts(config, template_name)
                        key = ( config.name, os.path.splitext(template_name)[0] )
                        self.message('Found template: {}; static files: {}'.format(key, scripts), 3)
                        script_map[key] = scripts

            for subdir in subdirs:
                recurse(subdir)

        recurse(template_root)
        return script_map


    def template_scripts(self, config, template_name):
        '''
        Returns a list of scripts used by the given template object AND its ancestors.

        This runs a ProviderRun on the given template (as if it were being displayed).
        This allows the WEBPACK_PROVIDERS to provide the JS files to us.
        '''
        dmp = apps.get_app_config('django_mako_plus')
        template_obj = dmp.engine.get_template_loader(config, create=True).get_mako_template(template_name, force=True)
        mako_context = create_mako_context(template_obj)
        inner_run = WebpackProviderRun(mako_context['self'])
        inner_run.run()
        scripts = []
        for tpl in inner_run.templates:
            for p in tpl.providers:
                if os.path.exists(p.absfilepath):
                    scripts.append(p.absfilepath)
        return scripts



###############################################################################
###   Specialized provider run for the above management command

class WebpackProviderRun(ProviderRun):
    SETTINGS_KEY = 'WEBPACK_PROVIDERS'

    def _get_template_inheritance(self):
        '''
        Normally, this returns a list of the template inheritance of tself, starting with the oldest ancestor.
        But for the webpack one, we just want the template itself, without any ancestors.
        This gives the static files for this exact template (not including all ancestors).
        '''
        return [ self.tself.template ]



from django.apps import apps
from django.conf import settings
from django.utils.module_loading import import_string
from collections import namedtuple
import io
import inspect
from uuid import uuid1
from ..template import template_inheritance
from ..util import qualified_name, b58enc


# initialize_providers() below creates a list of these to hold provider options from settings.py
# I can't keep the options inside the provider class itself because a given class can be listed
# more than once in settings.py (with different options).
ProviderClassInfo = namedtuple("ProviderClassInfo", [ 'cls', 'options' ])

# __init__() below creates a list of templates, each of which has a list of providers
# this named tuple adds a small amount of extra clarity to it.
# I could use a dict or OrderedDict, but I need order AND fast indexing
TemplateProviderList = namedtuple("TemplateProviderList", [ 'template', 'providers' ])


####################################################
###   Main runner for providers

class ProviderRun(object):
    '''A run through the providers for tself and its ancestors'''
    SETTINGS_KEY = 'CONTENT_PROVIDERS'
    CONTENT_PROVIDERS = []

    @classmethod
    def initialize_providers(cls):
        '''Initializes the providers (called from dmp app ready())'''
        dmp = apps.get_app_config('django_mako_plus')
        # regular content providers
        cls.CONTENT_PROVIDERS = []
        for provider_settings in dmp.options[cls.SETTINGS_KEY]:
            # import the class for this provider
            assert 'provider' in provider_settings, "Invalid entry in settings.py: CONTENT_PROVIDERS item must have 'provider' key"
            provider_cls = import_string(provider_settings['provider'])
            # combine options from all of its bases, then from settings.py
            options = {}
            for base in reversed(inspect.getmro(provider_cls)):
                options.update(getattr(base, 'DEFAULT_OPTIONS', {}))
            options.update(provider_settings)
            # add to the list
            if options['enabled']:
                cls.CONTENT_PROVIDERS.append(ProviderClassInfo(provider_cls, options))


    def __init__(self, tself, group=None):
        '''
        tself:              `self` object from a Mako template (available during rendering).
        group:              provider group to include (defaults to all groups if None)
        '''
        # a unique context id for this run
        self.uid = b58enc(uuid1().int)
        self.tself = tself
        self.request = tself.context.get('request')
        self.context = tself.context
        self.buffer = io.StringIO()

        # get the ProviderClassInfo objects that are used in this group
        group_pcis = [ pci for pci in self.CONTENT_PROVIDERS if group is None or pci.options['group'] == group ]

        # Create a map of template -> providers for this run
        #   {
        #       base.htm:      [ JsLinkProvider(), CssLinkProvider(), ... ]
        #       app_base.htm:  [ JsLinkProvider(), CssLinkProvider(), ... ]
        #       index.html:    [ JsLinkProvider(), CssLinkProvider(), ... ]
        #   }
        self.templates = []
        for tmpl in self._get_template_inheritance():
            tpl = TemplateProviderList(tmpl, [])
            for index, pci in enumerate(group_pcis):
                tpl.providers.append(pci.cls(self, tmpl, index, pci.options))
            self.templates.append(tpl)

    def _get_template_inheritance(self):
        '''Returns a list of the template inheritance of tself, starting with the oldest ancestor'''
        return reversed(list(template_inheritance(self.tself)))

    def run(self):
        '''Performs the run through the templates and their providers'''
        for tpl in self.templates:
            for provider in tpl.providers:
                provider.provide()

    def write(self, content):
        '''Provider instances use this to write to the buffer'''
        self.buffer.write(content)
        if settings.DEBUG:
            self.buffer.write('\n')

    def getvalue(self):
        '''Returns the buffer string'''
        return self.buffer.getvalue()
